(* $Id: godi_wishes.ml 678 2007-02-25 17:16:54Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_query
open Printf

type wish =
    [ `Build of string    (* package name *)
    | `Remove of string   (* package name *)
    | `Option of (string * string * string)  (* pkg name, opt name, opt val *)
    ]


let wishes_txt() =
  let lb = localbase() in
  Filename.concat (Filename.concat lb "build") "wishes.txt"
;;


exception Format_error


let esc_re = Str.regexp "[\000-\032%\127-\255]";;
let unesc_re = Str.regexp "%[0-9a-f][0-9a-f]";;

let escape_arg s =
  Str.global_substitute esc_re (fun u -> 
				  let m = Str.matched_string u in
				  sprintf "%%%02x" (Char.code m.[0])) s
;;


let unescape_arg s =
  Str.global_substitute 
    unesc_re
    (fun u ->
       let m = Str.matched_string u in
       let hex = String.copy "0x00" in
       hex.[2] <- m.[1];
       hex.[3] <- m.[2];
       String.make 1 (Char.chr (int_of_string hex)))
    s
;;


let space_re = Str.regexp "[ \t]+";;

let get_arguments line =
  let fields = Str.split space_re line in
  List.map unescape_arg fields
;;


let read_wishes ?(pref_user_error=false) () =
  let fn = wishes_txt() in
  if Sys.file_exists fn then (
    let f = 
      try 
	open_in_bin fn 
      with
	  Sys_error msg when pref_user_error ->
	    user_error("System error: " ^ msg)
    in
    let wl = ref [] in
    try
      while true do
	let line = input_line f in
	if line <> "" then (
	  let args = get_arguments line in
	  match args with
	      ["B"; pkg_name] ->
		wl := `Build pkg_name :: !wl
	    | ["R"; pkg_name] ->
		wl := `Remove pkg_name :: !wl
	    | ["O"; pkg_name; opt_name; opt_val] ->
		wl := `Option(pkg_name, opt_name, opt_val) :: !wl
	    | ["O"; pkg_name; opt_name] ->
		wl := `Option(pkg_name, opt_name, "") :: !wl
	    | _ -> 
		raise Format_error
	)
      done;
      assert false
    with
	End_of_file ->
	  close_in f;
	  List.rev !wl
      | Format_error ->
	  close_in f;
	  if pref_user_error then
	    user_error("File has wrong format: " ^ wishes_txt())
	  else
	    failwith ("File has wrong format: " ^ wishes_txt())
      | error ->
	  close_in f;
	  raise error
  )
  else
    []
;;


let write_wishes ?(pref_user_error=false) wl =
  let f = 
    try
      open_out_bin (wishes_txt()) 
    with
	Sys_error msg when pref_user_error ->
	  user_error("System error: " ^ msg)
  in
  ( try
      List.iter
	(function
	     `Build name ->
	       fprintf f "B %s\n" (escape_arg name)
	   | `Remove name ->
	       fprintf f "R %s\n" (escape_arg name)
	   | `Option(pkg_name,opt_name,opt_val) ->
	       fprintf f "O %s %s %s\n" 
	         (escape_arg pkg_name)
	         (escape_arg opt_name)
	         (escape_arg opt_val)
	)
	wl
    with
	error -> close_out f; raise error
  );
  close_out f
;;


let rec wishes_of_packages pl =
  match pl with
      [] -> []
    | pkg :: pl' ->
	let pkgobj = any_pkg pkg in
	( match pkgobj # user_status with
	      `Build -> 
		(`Build pkgobj#name) :: 
		(options pkg @ wishes_of_packages pl')
	    | `Remove ->
		(`Remove pkgobj#name) :: 
		wishes_of_packages pl'
	    | _ ->
		options pkg @ wishes_of_packages pl'
	)
and options pkg =
  match pkg with
      `Source_pkg pkgobj ->
	let user_opts = pkgobj#values_of_user_build_options in
	let curr_opts = pkgobj#values_of_build_options in
	List.flatten
	  (List.map2
	     (fun (uname,uval) (cname,cval) ->
		assert(uname = cname);
		if uval = cval then
		  []
		else
		  [`Option(pkgobj#name,uname,uval)])
	     user_opts
	     curr_opts)
	     
    | _ ->
	[]
;;


let apply_wishes_to_packages ?iteration pl wl =
  List.iter 
    (fun pkg -> (any_pkg pkg) # set_user_status `Keep) 
    pl;

  List.iter
    (function 
	 `Build name ->
	   ( try
	       let pkg =
		 List.find 
		   (function `Source_pkg p -> p#name = name | _ -> false)
		   pl in
	       let status =
		 match iteration with
		     None -> `Build
		   | Some i -> `Build_dep i in
	       (any_pkg pkg) # set_user_status status
	     with
		 Not_found -> ()
	   )
       | `Remove name ->
	   ( try
	       let pkg =
		 List.find 
		   (function `Installed_pkg p -> p#name = name | _ -> false)
		   pl in
	       let status =
		 match iteration with
		     None -> `Remove
		   | Some i -> `Remove_dep i in
	       (any_pkg pkg) # set_user_status status
	     with
		 Not_found -> ()
	   )
       | `Option(pkg_name,opt_name,opt_val) ->
	   ( try
	       let pkg =
		 List.find 
		   (function `Source_pkg p -> p#name = pkg_name | _ -> false)
		   pl in
	       let p = source_pkg pkg in
	       let opts =
		 List.map
		   (fun (n,v) ->
		      if n = opt_name then
			(n,opt_val)
		      else
			(n,v))
		   p#values_of_user_build_options in
	       p#set_user_build_options opts
	     with
		 Not_found -> ()
	   )
    )
    wl
;;


  
